#ifndef _DEVICE_H
#define _DEVICE_H

/*
   Devices will trigger specific sequences of events to perform the derired actions. For instance, a 
   simple action would be "Turn a light ON".  This will trigger the event to turn ON the device associated 
   with the light. " OPEN the window" is a more complex action, because it is not instant ant it takes 
   several steps. Make sure the CLOSE button isn't pressed, then press the OPEN button, and after a few 
   seconds stop pressing it. All of these sequence of events are different according to each device type.
*/

#include <WProgram.h>
#include "Event.h"

// Macros to simplify the creation of new devices
// They seem complicated but they are only creating unique identifiers for each address name
// For instance:  ADD_DEVICE_SWITCH( "Kitchen light", DEVICE_CONTROL_RADIO, IMAGE_LIGHT_ON, 0, 3);
// Creates the code P(dev_sw_nam_2_0_3) = "Kitchen light"; addDevice(..........)
// These unique identifiers are necessary to keep all the strings in progmem, saving precious RAM
// Then the addDevice function in called, passing the parameters in the right positions for each different device
// Some of the parameters are common to all of the devices:
// name - the name that is associated with the device
// control - the type of controller associated with the device
// image - reference image for the device
// address - address of the device on that control
#define ADD_DEVICE_SWITCH(name, control, image, state, address) P(dev_sw_nam_##control_##state_##address) = name; addDevice(dev_sw_nam_##control_##state_##address, DEVICE_TYPE_SWITCH, control, image, state, address, 255, 255);
// Sliders have 2 adresses, since they work in a more compelex way. One address makes it go up, another makes it go down
// slidetime - time it takes to change from one state to the other
#define ADD_DEVICE_SLIDER(name, control, image, state, address1, address2, slidetime) P(dev_sw_nam_##control##_##state##_##address1##_##address2) = name; addDevice(dev_sw_nam_##control##_##state##_##address1##_##address2, DEVICE_TYPE_SLIDER, control, image, state, address1, address2, slidetime);
// The timer is similar to the switch, but stays on for a specified number of minutes/seconds
#define ADD_DEVICE_TIMER(name, control, image, state, address, timerminutes, timerseconds) P(dev_sw_nam_##control##_##state##_##address) = name; addDevice(dev_sw_nam_##control##_##state##_##address, DEVICE_TYPE_TIMER, control, image,  state, address, timerminutes, timerseconds);

// Different types of devices implemented
#define DEVICE_TYPE_SWITCH 1
#define DEVICE_TYPE_SLIDER 2
#define DEVICE_TYPE_TIMER 3

// Different types of control implemented
// Defined in Event.h, this here is just a reminder
//#define DEVICE_CONTROL_SHIFT_REGISTER      1
//#define DEVICE_CONTROL_CHACON              2
//#define DEVICE_CONTROL_IR                  3
//#define DEVICE_CONTROL_DIRECT_PIN          4

// The same structure is used for all device types, to keep the code shorter and save space
// Many attributes are common, but some are individual to specific devices. Those are stored
// in the byte array "data", and processed differently when dealing with eaach device
/*
typedef struct
{
    const prog_uchar* name;    // Name of the device
    byte type;                 // Type of the device
    byte control;              // Type of control used on the device  
    byte image;                // First image associated with the device
    int pos_x;                 // X position of the device in the blueprint
    int pos_y;                 // Y position of the device in the blueprint
    byte data[4];              // Generic data, different for each device
} tDevice, *pDevice;

// Ordered Linked List containing all the devices active in the system
OrderedList <tDevice> devices(compareDevices, destroyDevice); 

*/

// Adds a new device to the device list
// d1 d2 d3 d4 are generic bytes that have different meanings depending on the device type
void addDevice(const prog_uchar* dev_name, byte type, byte control, byte image, byte d1, byte d2, byte d3, byte d4)
{
    pDevice new_device = (pDevice) malloc(sizeof(tDevice));
    new_device-> name = dev_name;
    new_device->type = type;
    new_device->control = control;
    new_device->image = image;
    new_device->pos_x  = -1;
    new_device->pos_y  = -1;
    new_device->data[0]  = d1;
    new_device->data[1]  = d2;
    new_device->data[2]  = d3;
    new_device->data[3]  = d4;    
    devices.addElement(new_device);    
}

// Returns the device type of a specific device
byte getDeviceType(byte device_number)
{
    return devices.getPosition(device_number)->type;
}

// Returns the device status of a specific device
byte getDeviceStatus(byte device_number)
{
    return devices.getPosition(device_number)->data[0];
}

// Returns the image of a specific device
// This image will be different according to the current state of the device
byte getDeviceImage(byte device_number, byte device_status)
{
    byte img = devices.getPosition(device_number)->image;
    if (device_status == 0 ) img++;
    return img;
}

// This function is called to change the status of a device, no matter if it's a direct order
// or a scheduled order
void setDeviceStatus(byte device_num, byte new_status)
{
    pDevice dev = devices.getPosition(device_num);
    dev->data[0] = new_status;
    #ifdef DEBUG_DEVICES
    Serial.print("Setting device:");  Serial.print(device_num, DEC);  
    Serial.print("  type:"); Serial.print(dev->type, DEC);
    Serial.print("  value:"); Serial.println(new_status, DEC);
    #endif    
    long end_time = getDaySeconds() + dev->data[3];
    // This is the part where each device data need to be interpreted 
    switch(dev->type)
    {
        // If the device is a SWITCH, simply sets the new status
        case DEVICE_TYPE_SWITCH: sendOrder(dev->control, dev->data[1], new_status, device_num, new_status);
                                 break;
        // If the device is a SLIDER, ir requires severeal steps
        // Remember a slider has two controls, let's call them UP and DOWN
                                 // First it removes all the events associated with that slider. This is important because
                                 // this funcion can be called when the slider is already moving
        case DEVICE_TYPE_SLIDER: removeEvent(dev->control, dev->data[1]); removeEvent(dev->control, dev->data[2]);
                                 // Then it deactivates control of the opposite direction, and activates the control of
                                 // desired direction. For instance, if we want to go UP, first we make sure the DOWN
                                 // switch is disabled, only then we enable the UP switch
                                 sendOrder(dev->control, (new_status ? dev->data[2] : dev->data[1]), 0, device_num, new_status); 
                                 sendOrder(dev->control, (new_status ? dev->data[1] : dev->data[2]), 1, device_num, new_status); 
                                 // After the slider is moving in the right direction, we schedule an event to make it stop
                                 // moving after the desired time
                                 addEvent(end_time, dev->control, (new_status ? dev->data[1] : dev->data[2]), 0, device_num, new_status);
                                 break;
        // If the device is a TIMER, it also requires several steps
                                 // First it removes all the events associated with that timer
        case DEVICE_TYPE_TIMER:  removeEvent(dev->control, dev->data[1]);
                                 // Then it turns the timer ON
                                 sendOrder(dev->control, dev->data[1] , 1, device_num, 1); 
                                 // Finally, schedules an event to set the timer OFF after the desired time
                                 addEvent(end_time, dev->control, dev->data[1], 0, device_num, 0); 
                                 break;
    }    
}

// Gets the X position of the device in the blueprint of the house
float getDeviceXPos(byte device_number)
{
    return int2Float(devices.getPosition(device_number)->pos_x);
}

// Gets the Y position of the device in the blueprint of the house
float getDeviceYPos(byte device_number)
{
    return int2Float(devices.getPosition(device_number)->pos_y);
}

// Sets the X, Y position of the device in the blueprint of the house
void setDevicePosition(byte device_number, float xpos, float ypos)
{
    pDevice device = devices.getPosition(device_number);
    device->pos_x  = float2Int(xpos);
    device->pos_y  = float2Int(ypos);
    EEPROM_writeAnything(0+(device_number-1)*4, device->pos_x);
    EEPROM_writeAnything(2+(device_number-1)*4, device->pos_y);
}

// Reads the X, Y position of the device in the blueprint of the house from the Arduino Flash memory
void deviceReadPosition(byte device_number)
{
    pDevice device = devices.getPosition(device_number);
    EEPROM_readAnything(0+(device_number-1)*4, device->pos_x);
    EEPROM_readAnything(2+(device_number-1)*4, device->pos_y);
}

// Loads all the device positions from the Arduino Flash memory
void loadDevicePositions()
{
    for (int i=1; i<=devices.count(); i++)
    {
        deviceReadPosition(i);
    }
}

// Resets all the device positions from the Arduino Flash memory
void resetDevicePositions()
{
    for (int i=1; i<=devices.count(); i++)
    {
        setDevicePosition(i, -1, -1);
    }
}

// Returns a pointer to the device name, stored in progmem
const prog_uchar* getDeviceName(int i)
{
    return devices.getPosition(i)->name;
}

// Returns the nest device number which doesn't have a X,Y position defined
byte nextDeviceWithNoPosition()
{
    for (byte i=1; i<=devices.count(); i++)
    {
        if (getDeviceXPos(i) < 0)
            return i;
    }
    return 0;
}


#endif
